@@ -90,6 +90,8 @@ struct layer {
   int32_t tiles_across;
   int32_t tiles_down;
 
+  double downsample;
+
   int32_t scale_denom;
 
   // how much extra we might need to read to get all relevant tiles?
@@ -622,8 +624,6 @@ static void read_tile(openslide_t *osr,
   int64_t tileindex = tile_y * l->tiles_across + tile_x;
   struct tile *requested_tile = (struct tile *) g_hash_table_lookup(l->tiles, &tileindex);
 
-  bool cachemiss;
-
   if (!requested_tile) {
     //    g_debug("no tile at index %" G_GINT64_FORMAT, tileindex);
     return;
@@ -634,22 +634,30 @@ static void read_tile(openslide_t *osr,
   }
 
   // get the jpeg data, possibly from cache
+  struct _openslide_cache_entry *cache_entry;
   g_mutex_lock(data->cache_mutex);
   uint32_t *tiledata = (uint32_t *) _openslide_cache_get(cache,
 							 requested_tile->jpegno,
 							 requested_tile->tileno,
-							 layer);
+							 layer,
+							 &cache_entry);
   g_mutex_unlock(data->cache_mutex);
   int tw = requested_tile->jpeg->tile_width / l->scale_denom;
   int th = requested_tile->jpeg->tile_height / l->scale_denom;
 
-  cachemiss = !tiledata;
   if (!tiledata) {
     tiledata = read_from_one_jpeg(osr,
 				  requested_tile->jpeg,
 				  requested_tile->tileno,
 				  l->scale_denom,
 				  tw, th);
+
+    g_mutex_lock(data->cache_mutex);
+    _openslide_cache_put(cache, requested_tile->jpegno, requested_tile->tileno, layer,
+			 tiledata,
+			 tw * th * 4,
+			 &cache_entry);
+    g_mutex_unlock(data->cache_mutex);
   }
 
   // draw it
@@ -724,14 +732,8 @@ static void read_tile(openslide_t *osr,
   */
 
 
-  // put into cache last, because the cache can free this tile
-  if (cachemiss) {
-    g_mutex_lock(data->cache_mutex);
-    _openslide_cache_put(cache, requested_tile->jpegno, requested_tile->tileno, layer,
-			 tiledata,
-			 tw * th * 4);
-    g_mutex_unlock(data->cache_mutex);
-  }
+  // done with the cache entry, release it
+  _openslide_cache_entry_unref(cache_entry);
 }
 
 
@@ -1148,6 +1150,7 @@ void _openslide_add_jpeg_ops(openslide_t *osr,
     struct layer *new_l = g_slice_new0(struct layer);
     new_l->tiles_across = old_l->tiles_across;
     new_l->tiles_down = old_l->tiles_down;
+    new_l->downsample = old_l->downsample;
     new_l->scale_denom = 1;
     new_l->pixel_w = old_l->layer_w;
     new_l->pixel_h = old_l->layer_h;
@@ -1181,6 +1184,7 @@ void _openslide_add_jpeg_ops(openslide_t *osr,
       sd_l->tiles = g_hash_table_ref(new_l->tiles);
       sd_l->tiles_across = new_l->tiles_across;
       sd_l->tiles_down = new_l->tiles_down;
+      sd_l->downsample = new_l->downsample * scale_denom;
       sd_l->extra_tiles_top = new_l->extra_tiles_top;
       sd_l->extra_tiles_bottom = new_l->extra_tiles_bottom;
       sd_l->extra_tiles_left = new_l->extra_tiles_left;
@@ -1221,6 +1225,10 @@ void _openslide_add_jpeg_ops(openslide_t *osr,
   // populate the layer_count
   osr->layer_count = g_hash_table_size(expanded_layers);
 
+  // allocate downsample array
+  g_assert(osr->downsamples == NULL);
+  osr->downsamples = g_new(double, osr->layer_count);
+
   // load into data array
   data->layers = g_new(struct layer, g_hash_table_size(expanded_layers));
   GList *tmp_list = layer_keys;
@@ -1236,6 +1244,9 @@ void _openslide_add_jpeg_ops(openslide_t *osr,
     struct layer *dest = data->layers + i;
     *dest = *l;    // shallow copy
 
+    // set downsample
+    osr->downsamples[i] = l->downsample;
+
     // manually free some things, because of that shallow copy
     g_hash_table_steal(expanded_layers, tmp_list->data);
     _openslide_int64_free(tmp_list->data);  // key
