@@ -1,7 +1,7 @@
 /*
  *  OpenSlide, a library for reading whole slide image files
  *
- *  Copyright (c) 2007-2011 Carnegie Mellon University
+ *  Copyright (c) 2007-2012 Carnegie Mellon University
  *  Copyright (c) 2011 Google, Inc.
  *  All rights reserved.
  *
@@ -110,6 +110,8 @@ static const char KEY_IMAGE_CONCAT_FACTOR[] = "IMAGE_CONCAT_FACTOR";
   } while(0)
 
 struct slide_zoom_level_section {
+  int concat_exponent;
+
   double overlap_x;
   double overlap_y;
 
@@ -119,19 +121,14 @@ struct slide_zoom_level_section {
   int tile_h;
 };
 
-struct mirax_hier_page_entry {
-  int32_t x;
-  int32_t y;
-  int32_t offset;
-  int32_t length;
-  int32_t fileno;
-  int zoom_level;
-};
-
-struct mirax_nonhier_page_entry {
-  int32_t offset;
-  int32_t length;
-  int32_t fileno;
+// see comments in _openslide_try_mirax()
+struct slide_zoom_level_params {
+  int tile_concat;
+  int tile_count_divisor;
+  int subtiles_per_jpeg_tile;
+  int positions_per_subtile;
+  double subtile_w;
+  double subtile_h;
 };
 
 static char *read_string_from_file(FILE *f, int len) {
@@ -186,15 +183,8 @@ static bool read_nonhier_record(FILE *f,
     return false;
   }
 
-  // jump to start of interesting data
-  //  g_debug("seek %d", ptr);
-  if (fseeko(f, ptr, SEEK_SET) == -1) {
-    g_warning("Cannot seek to start of nonhier data");
-    return false;
-  }
-
   // seek to record pointer
-  if (fseeko(f, recordno * 4, SEEK_CUR) == -1) {
+  if (fseeko(f, ptr + 4 * recordno, SEEK_SET) == -1) {
     g_warning("Cannot seek to nonhier record pointer %d", recordno);
     return false;
   }
@@ -307,6 +297,75 @@ static void insert_subtile(GHashTable *tiles, int32_t jpeg_number,
   }
 }
 
+// given the coordinates of a subtile, compute its layer 0 pixel coordinates.
+// return false if none of the camera positions within the subtile are
+// active.
+static bool get_subtile_position(int32_t *tile_positions,
+                                 GHashTable *active_positions,
+                                 const struct slide_zoom_level_params *slide_zoom_level_params,
+                                 struct _openslide_jpeg_layer **layers,
+                                 int tiles_across,
+                                 int image_divisions,
+                                 int zoom_level, int xx, int yy,
+                                 int *pos0_x, int *pos0_y)
+{
+  const struct slide_zoom_level_params *lp = slide_zoom_level_params +
+      zoom_level;
+
+  const int tile0_w = layers[0]->raw_tile_width;
+  const int tile0_h = layers[0]->raw_tile_height;
+
+  // camera position coordinates
+  int xp = xx / image_divisions;
+  int yp = yy / image_divisions;
+  int tp = yp * (tiles_across / image_divisions) + xp;
+  //g_debug("xx %d, yy %d, xp %d, yp %d, tp %d, spp %d, sc %d, tile0: %d %d subtile: %g %g", xx, yy, xp, yp, tp, subtiles_per_position, lp->subtiles_per_jpeg_tile, tile0_w, tile0_h, lp->subtile_w, lp->subtile_h);
+
+  *pos0_x = tile_positions[tp * 2] +
+      tile0_w * (xx - xp * image_divisions);
+  *pos0_y = tile_positions[(tp * 2) + 1] +
+      tile0_h * (yy - yp * image_divisions);
+
+  // ensure only active positions (those present at zoom level 0) are
+  // processed at higher zoom levels
+  if (zoom_level == 0) {
+    // If the layer 0 concat factor <= image_divisions, we can simply mark
+    // active any position with a corresponding level 0 tile.
+    //
+    // If the concat factor is larger, then active and inactive positions
+    // can be merged into the same tile, and we can no longer tell which
+    // subtiles can be skipped at higher zoom levels.  Sometimes such
+    // positions have coordinates (0, 0) in the tile_positions map; we can
+    // at least filter out these, and we must because such positions break
+    // the JPEG backend's range search.  Assume that only position (0, 0)
+    // can be at pixel (0, 0).
+    if (tile_positions[tp * 2] == 0 && tile_positions[tp * 2 + 1] == 0 &&
+        (xp != 0 || yp != 0)) {
+      return false;
+    }
+
+    int *key = g_new(int, 1);
+    *key = tp;
+    g_hash_table_insert(active_positions, key, NULL);
+    return true;
+
+  } else {
+    // make sure at least one of the positions within this subtile is active
+    for (int ypp = yp; ypp < yp + lp->positions_per_subtile; ypp++) {
+      for (int xpp = xp; xpp < xp + lp->positions_per_subtile; xpp++) {
+        int tpp = ypp * (tiles_across / image_divisions) + xpp;
+        if (g_hash_table_lookup_extended(active_positions, &tpp, NULL, NULL)) {
+          //g_debug("accept tile: level %d xp %d yp %d xpp %d ypp %d", zoom_level, xp, yp, xpp, ypp);
+          return true;
+        }
+      }
+    }
+
+    //g_debug("skip tile: level %d positions %d xp %d yp %d", zoom_level, lp->positions_per_subtile, xp, yp);
+    return false;
+  }
+}
+
 static bool process_hier_data_pages_from_indexfile(FILE *f,
 						   int64_t seek_location,
 						   int datafile_count,
@@ -317,6 +376,7 @@ static bool process_hier_data_pages_from_indexfile(FILE *f,
 						   int tiles_across,
 						   int tiles_down,
 						   int image_divisions,
+						   const struct slide_zoom_level_params *slide_zoom_level_params,
 						   int32_t *tile_positions,
 						   GList **jpegs_list,
 						   struct _openslide_hash *quickhash1) {
@@ -330,6 +390,8 @@ static bool process_hier_data_pages_from_indexfile(FILE *f,
 
   for (int zoom_level = 0; zoom_level < zoom_levels; zoom_level++) {
     struct _openslide_jpeg_layer *l = layers[zoom_level];
+    const struct slide_zoom_level_params *lp = slide_zoom_level_params +
+        zoom_level;
     int32_t ptr;
 
     //    g_debug("reading zoom_level %d", zoom_level);
@@ -461,36 +523,21 @@ static bool process_hier_data_pages_from_indexfile(FILE *f,
 
 	*jpegs_list = g_list_prepend(*jpegs_list, jpeg);
 
-
-	// see comments elsewhere in this file
-	const int tile_concat = 1 << zoom_level;
-	const int tile_count_divisor = MIN(tile_concat, image_divisions);
-	const int subtiles_per_jpeg_tile = MAX(1, tile_concat / image_divisions);
-	const double subtile_w = (double) jpeg->w / subtiles_per_jpeg_tile;
-	const double subtile_h = (double) jpeg->h / subtiles_per_jpeg_tile;
-
-	const int tile0_w = layers[0]->raw_tile_width;
-	const int tile0_h = layers[0]->raw_tile_height;
-
-	// subtile_count: how many subtiles in a JPEG tile (one dimension)? this is constant
-	//                for the first few levels, depending on image_divisions
-	const int subtile_count = MAX(1, tile_concat / image_divisions);
-
 	/*
-	g_debug("tile_concat: %d, subtile_count: %d",
-		tile_concat, subtile_count);
+	g_debug("tile_concat: %d, subtiles_per_jpeg_tile: %d",
+		lp->tile_concat, lp->subtiles_per_jpeg_tile);
 	g_debug("found %d %d from file", x, y);
 	*/
 
 
-	// start processing 1 JPEG tile into subtile_count^2 subtiles
-	for (int yi = 0; yi < subtile_count; yi++) {
+	// start processing 1 JPEG tile into subtiles_per_jpeg_tile^2 subtiles
+	for (int yi = 0; yi < lp->subtiles_per_jpeg_tile; yi++) {
 	  int yy = y + (yi * image_divisions);
 	  if (yy >= tiles_down) {
 	    break;
 	  }
 
-	  for (int xi = 0; xi < subtile_count; xi++) {
+	  for (int xi = 0; xi < lp->subtiles_per_jpeg_tile; xi++) {
 	    int xx = x + (xi * image_divisions);
 	    if (xx >= tiles_across) {
 	      break;
@@ -498,43 +545,36 @@ static bool process_hier_data_pages_from_indexfile(FILE *f,
 
 	    // xx and yy are the tile coordinates in level0 space
 
-	    // look up the tile position, stored in 24.8 fixed point
-	    int xp = xx / image_divisions;
-	    int yp = yy / image_divisions;
-	    int tp = yp * (tiles_across / image_divisions) + xp;
-	    //g_debug("xx %d, yy %d, xp %d, yp %d, tp %d, spp %d, sc %d, tile0: %d %d subtile: %g %g", xx, yy, xp, yp, tp, subtiles_per_position, subtile_count, tile0_w, tile0_h, subtile_w, subtile_h);
-
-	    if (zoom_level == 0) {
-	      // if the zoom level is 0, then mark this position as active
-	      int *key = g_new(int, 1);
-	      *key = tp;
-	      g_hash_table_insert(active_positions, key, NULL);
-	    } else {
-	      // make sure we have an active position for this tile
-	      if (!g_hash_table_lookup_extended(active_positions, &tp, NULL, NULL)) {
-		continue;
-	      }
-	    }
-
 	    // position in layer 0
-	    const double pos0_x = ((double) tile_positions[tp * 2]) / 256.0 +
-	      tile0_w * (xx - xp * image_divisions);
-	    const double pos0_y = ((double) tile_positions[(tp * 2) + 1]) / 256.0 +
-	      tile0_h * (yy - yp * image_divisions);
+            int pos0_x;
+            int pos0_y;
+            if (!get_subtile_position(tile_positions,
+                                      active_positions,
+                                      slide_zoom_level_params,
+                                      layers,
+                                      tiles_across,
+                                      image_divisions,
+                                      zoom_level,
+                                      xx, yy,
+                                      &pos0_x, &pos0_y)) {
+              // no such position
+              continue;
+            }
 
 	    // position in this layer
-	    const double pos_x = pos0_x / tile_concat;
-	    const double pos_y = pos0_y / tile_concat;
+	    const double pos_x = ((double) pos0_x) / lp->tile_concat;
+	    const double pos_y = ((double) pos0_y) / lp->tile_concat;
 
-	    //g_debug("pos0: %g %g, pos: %g %g", pos0_x, pos0_y, pos_x, pos_y);
+	    //g_debug("pos0: %d %d, pos: %g %g", pos0_x, pos0_y, pos_x, pos_y);
 
 	    insert_subtile(l->tiles, jpeg_number,
 			   pos_x, pos_y,
-			   subtile_w * xi, subtile_h * yi,
-			   subtile_w, subtile_h,
+			   lp->subtile_w * xi, lp->subtile_h * yi,
+			   lp->subtile_w, lp->subtile_h,
 			   l->tile_advance_x, l->tile_advance_y,
-			   x / tile_count_divisor + xi, y / tile_count_divisor + yi,
-			   tiles_across / tile_count_divisor,
+			   x / lp->tile_count_divisor + xi,
+			   y / lp->tile_count_divisor + yi,
+			   tiles_across / lp->tile_count_divisor,
 			   zoom_level);
 	  }
 	}
@@ -555,7 +595,8 @@ static bool process_hier_data_pages_from_indexfile(FILE *f,
 }
 
 static int32_t *read_slide_position_file(const char *dirname, const char *name,
-					 int64_t size, int64_t offset) {
+					 int64_t size, int64_t offset,
+					 int level_0_tile_concat) {
   char *tmp = g_build_filename(dirname, name, NULL);
   FILE *f = _openslide_fopen(tmp, "rb");
   g_free(tmp);
@@ -577,23 +618,23 @@ static int32_t *read_slide_position_file(const char *dirname, const char *name,
   //  g_debug("tile positions count: %d", count);
 
   for (int i = 0; i < count; i++) {
-    // read 2 numbers, then a null
+    // read flag byte, then 2 numbers
+    int zz = getc(f);
+
     int32_t x;
     int32_t y;
     bool x_ok = read_le_int32_from_file_with_result(f, &x);
     bool y_ok = read_le_int32_from_file_with_result(f, &y);
 
-    int zz = getc(f);
-
-    if (!x_ok || !y_ok || (zz == EOF) || ((zz != 0) && (zz != 255))) {
-      g_warning("Error while reading slide position file");
+    if (zz == EOF || !x_ok || !y_ok || (zz & 0xfe)) {
+      g_warning("Error while reading slide position file (%d)", zz);
       fclose(f);
       g_free(result);
       return NULL;
     }
 
-    result[i * 2] = x;
-    result[(i * 2) + 1] = y;
+    result[i * 2] = x * level_0_tile_concat;
+    result[(i * 2) + 1] = y * level_0_tile_concat;
   }
 
   fclose(f);
@@ -628,6 +669,7 @@ static bool process_indexfile(const char *uuid,
 			      int tiles_x,
 			      int tiles_y,
 			      int image_divisions,
+			      const struct slide_zoom_level_params *slide_zoom_level_params,
 			      FILE *indexfile,
 			      struct _openslide_jpeg_layer **layers,
 			      int *file_count_out,
@@ -700,7 +742,8 @@ static bool process_indexfile(const char *uuid,
     slide_positions = read_slide_position_file(dirname,
 					       datafile_names[slide_position_fileno],
 					       slide_position_size,
-					       slide_position_offset);
+					       slide_position_offset,
+					       slide_zoom_level_params[0].tile_concat);
   } else {
     // no position map available and we know overlap is 0, fill in our own
     // values based on the known tile size.
@@ -709,8 +752,8 @@ static bool process_indexfile(const char *uuid,
 
     slide_positions = g_new(int, ntiles * 2);
     for (int i = 0; i < ntiles; i++) {
-      slide_positions[(i * 2)]     = (i % tiles_x) * tile0_w * 256;
-      slide_positions[(i * 2) + 1] = (i / tiles_x) * tile0_h * 256;
+      slide_positions[(i * 2)]     = (i % tiles_x) * tile0_w;
+      slide_positions[(i * 2) + 1] = (i / tiles_x) * tile0_h;
     }
   }
   if (!slide_positions) {
@@ -782,13 +825,6 @@ static bool process_indexfile(const char *uuid,
     goto DONE;
   }
 
-  // jump to start of interesting data
-  //  g_debug("seek %d", ptr);
-  if (fseeko(indexfile, ptr, SEEK_SET) == -1) {
-    g_warning("Cannot seek to start of interesting data");
-    goto DONE;
-  }
-
   // read these pages in
   if (!process_hier_data_pages_from_indexfile(indexfile,
 					      ptr,
@@ -800,6 +836,7 @@ static bool process_indexfile(const char *uuid,
 					      tiles_x,
 					      tiles_y,
 					      image_divisions,
+					      slide_zoom_level_params,
 					      slide_positions,
 					      &jpegs_list,
 					      quickhash1)) {
@@ -1006,6 +1043,7 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
   char *key_slide_zoom_level_count = NULL;
   char **slide_zoom_level_section_names = NULL;
   struct slide_zoom_level_section *slide_zoom_level_sections = NULL;
+  struct slide_zoom_level_params *slide_zoom_level_params = NULL;
 
   int datafile_count = 0;
   char **datafile_names = NULL;
@@ -1017,6 +1055,8 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
   int64_t base_w = 0;
   int64_t base_h = 0;
 
+  int total_concat_exponent = 0;
+
   // start reading
 
   // verify filename
@@ -1177,6 +1217,9 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
       goto FAIL;
     }
 
+    READ_KEY_OR_FAIL(hs->concat_exponent, slidedat, group,
+		     KEY_IMAGE_CONCAT_FACTOR,
+		     integer, "Can't read image concat exponent");
     READ_KEY_OR_FAIL(hs->overlap_x, slidedat, group, KEY_OVERLAP_X,
 		     double, "Can't read overlap X");
     READ_KEY_OR_FAIL(hs->overlap_y, slidedat, group, KEY_OVERLAP_Y,
@@ -1188,6 +1231,11 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
     READ_KEY_OR_FAIL(hs->tile_h, slidedat, group, KEY_DIGITIZER_HEIGHT,
 		     integer, "Can't read tile height");
 
+    if (i == 0) {
+      NON_NEGATIVE_OR_FAIL(hs->concat_exponent);
+    } else {
+      POSITIVE_OR_FAIL(hs->concat_exponent);
+    }
     POSITIVE_OR_FAIL(hs->tile_w);
     POSITIVE_OR_FAIL(hs->tile_h);
 
@@ -1206,19 +1254,6 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
     }
     g_free(tmp);
     tmp = NULL;
-
-    // verify IMAGE_CONCAT_FACTOR == 1 for all but the first layer
-    int ic_factor;
-    READ_KEY_OR_FAIL(ic_factor, slidedat, group, KEY_IMAGE_CONCAT_FACTOR,
-		     integer, "Can't read image concat factor");
-    if ((i == 0) && (ic_factor != 0)) {
-      g_warning("Level 0 has non-zero image concat factor: %d", ic_factor);
-      goto FAIL;
-    }
-    if ((i != 0) && (ic_factor != 1)) {
-      g_warning("Level %d has non-unity image concat factor: %d", i, ic_factor);
-      goto FAIL;
-    }
   }
 
 
@@ -1336,40 +1371,66 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
 
   // set up layer dimensions and such
   layers = g_new(struct _openslide_jpeg_layer *, zoom_levels);
+  slide_zoom_level_params = g_new(struct slide_zoom_level_params, zoom_levels);
+  total_concat_exponent = 0;
   for (int i = 0; i < zoom_levels; i++) {
     // one jpeg layer per zoom level
     struct _openslide_jpeg_layer *l = g_slice_new0(struct _openslide_jpeg_layer);
     layers[i] = l;
     struct slide_zoom_level_section *hs = slide_zoom_level_sections + i;
+    struct slide_zoom_level_params *lp = slide_zoom_level_params + i;
 
     // tile_concat: number of tiles concatenated from the original in one dimension
-    const int tile_concat = 1 << i;
+    total_concat_exponent += hs->concat_exponent;
+    lp->tile_concat = 1 << total_concat_exponent;
 
-    // subtiles_per_position: for this zoom, how many subtiles (in one dimension)
-    //                        come from a single photo?
-    const int subtiles_per_position = MAX(1, image_divisions / tile_concat);
+    // positions_per_jpeg_tile: for this zoom, how many camera positions
+    //                          are represented in a JPEG tile?
+    //                          this is constant for the first few levels,
+    //                          depending on image_divisions
+    const int positions_per_jpeg_tile = MAX(1, lp->tile_concat / image_divisions);
 
-    // tile_count_divisor: as we record levels, we would prefer to shrink the
-    //                     number of tiles, but keep the tile size constant,
-    //                     but this only works until we encounter JPEG tiles
-    //                     with more than one source photo, in which case
-    //                     the tile count bottoms out and we instead shrink
-    //                     the advances
-    const int tile_count_divisor = MIN(tile_concat, image_divisions);
+    if (position_nonhier_offset != -1) {
+      // tile_count_divisor: as we record levels, we would prefer to shrink the
+      //                     number of tiles, but keep the tile size constant,
+      //                     but this only works until we encounter JPEG tiles
+      //                     with more than one source photo, in which case
+      //                     the tile count bottoms out and we instead shrink
+      //                     the advances
+      lp->tile_count_divisor = MIN(lp->tile_concat, image_divisions);
 
-    // subtiles_per_jpeg_tile: for this zoom, how many subtiles in a JPEG tile?
-    const int subtiles_per_jpeg_tile = MAX(1, tile_concat / image_divisions);
+      // subtiles_per_jpeg_tile: for this zoom, how many subtiles in a JPEG tile?
+      //                         this is constant for the first few levels,
+      //                         depending on image_divisions
+      lp->subtiles_per_jpeg_tile = positions_per_jpeg_tile;
+
+      // positions_per_subtile: for this zoom, how many camera positions
+      //                        are represented in a subtile?
+      lp->positions_per_subtile = 1;
+
+    } else {
+      // no position file and no overlaps, so we can skip subtile processing
+      // for better performance
+
+      lp->tile_count_divisor = lp->tile_concat;
+      lp->subtiles_per_jpeg_tile = 1;
+      lp->positions_per_subtile = positions_per_jpeg_tile;
+    }
+
+    lp->subtile_w = (double) hs->tile_w / lp->subtiles_per_jpeg_tile;
+    lp->subtile_h = (double) hs->tile_h / lp->subtiles_per_jpeg_tile;
 
     l->tiles = _openslide_jpeg_create_tiles_table();
-    l->layer_w = base_w / tile_concat;  // tile_concat is powers of 2
-    l->layer_h = base_h / tile_concat;
-    l->tiles_across = tiles_x / tile_count_divisor;
-    l->tiles_down = tiles_y / tile_count_divisor;
+    l->layer_w = base_w / lp->tile_concat;  // tile_concat is powers of 2
+    l->layer_h = base_h / lp->tile_concat;
+    l->tiles_across = tiles_x / lp->tile_count_divisor;
+    l->tiles_down = tiles_y / lp->tile_count_divisor;
     l->raw_tile_width = hs->tile_w;  // raw JPEG size
     l->raw_tile_height = hs->tile_h;
 
-    double subtile_w = (double) hs->tile_w / subtiles_per_jpeg_tile;
-    double subtile_h = (double) hs->tile_h / subtiles_per_jpeg_tile;
+    // subtiles_per_position: for this zoom, how many subtiles (in one dimension)
+    //                        come from a single photo?
+    const int subtiles_per_position = MAX(1, image_divisions / lp->tile_concat);
 
     // use a fraction of the overlap, so that our tile correction will flip between
     // positive and negative values typically (in case image_divisions=2)
@@ -1377,10 +1438,16 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
 
     // overlaps are concatenated within physical tiles, so our virtual tile
     // size must shrink, once we hit image_divisions
-    l->tile_advance_x = subtile_w - ((double) hs->overlap_x / (double) subtiles_per_position);
-    l->tile_advance_y = subtile_h - ((double) hs->overlap_y / (double) subtiles_per_position);
+    l->tile_advance_x = lp->subtile_w - ((double) hs->overlap_x /
+        (double) subtiles_per_position);
+    l->tile_advance_y = lp->subtile_h - ((double) hs->overlap_y /
+        (double) subtiles_per_position);
 
-    //g_debug("layer %d tile advance %.10g %.10g, dim %" G_GINT64_FORMAT " %" G_GINT64_FORMAT ", tiles %d %d, rawtile %d %d, subtile %g %g, tile_concat %d, tile_count_divisor %d", i, l->tile_advance_x, l->tile_advance_y, l->layer_w, l->layer_h, l->tiles_across, l->tiles_down, l->raw_tile_width, l->raw_tile_height, subtile_w, subtile_h, tile_concat, tile_count_divisor);
+    // override downsample.  layer 0 is defined to have a downsample of 1.0,
+    // irrespective of its concat_exponent
+    l->downsample = lp->tile_concat / slide_zoom_level_params[0].tile_concat;
+
+    //g_debug("layer %d tile advance %.10g %.10g, dim %" G_GINT64_FORMAT " %" G_GINT64_FORMAT ", tiles %d %d, rawtile %d %d, subtile %g %g, tile_concat %d, tile_count_divisor %d, positions_per_subtile %d", i, l->tile_advance_x, l->tile_advance_y, l->layer_w, l->layer_h, l->tiles_across, l->tiles_down, l->raw_tile_width, l->raw_tile_height, lp->subtile_w, lp->subtile_h, lp->tile_concat, lp->tile_count_divisor, lp->positions_per_subtile);
   }
 
   // load the position map and build up the tiles, using subtiles
@@ -1398,6 +1465,7 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
 			 zoom_levels,
 			 tiles_x, tiles_y,
 			 image_divisions,
+			 slide_zoom_level_params,
 			 indexfile,
 			 layers,
 			 &num_jpegs, &jpegs,
@@ -1415,17 +1483,6 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
 
   _openslide_add_jpeg_ops(osr, num_jpegs, jpegs, zoom_levels, layers);
 
-  // override downsamples
-  if (osr) {
-    osr->downsamples = g_new(double, osr->layer_count);
-    double downsample = 1.0;
-
-    for (int32_t i = 0; i < osr->layer_count; i++) {
-      osr->downsamples[i] = downsample;
-      downsample *= 2.0;
-    }
-  }
-
   success = true;
   goto DONE;
 
@@ -1451,6 +1508,7 @@ bool _openslide_try_mirax(openslide_t *osr, const char *filename,
   g_strfreev(datafile_names);
   g_strfreev(slide_zoom_level_section_names);
   g_free(slide_zoom_level_sections);
+  g_free(slide_zoom_level_params);
   g_free(key_slide_zoom_level_name);
   g_free(key_slide_zoom_level_count);
 
